# distutils: language = c++

from libcpp.deque cimport deque
from libcpp.vector cimport vector
from libcpp.atomic cimport atomic
from libc.stdio cimport printf
from libc.stdlib cimport rand
from posix.unistd cimport sysconf
from runtime.pthreads cimport *
from runtime.semaphore cimport *

cdef extern from "<unistd.h>" nogil:
    enum: _SC_NPROCESSORS_ONLN # Seems to not be included in "posix.unistd".


cdef cypclass Scheduler nolock
cdef cypclass Worker nolock

# The 'inline' qualifier on this function is a hack to convince Cython to allow a definition in a .pxd file.
# The C compiler will dismiss it because we pass the function pointer to create a thread which prevents inlining.
cdef inline void * worker_function(void * arg) nogil:
    worker = <Worker> arg
    sch = worker.scheduler
    cdef int num_remaining_queues
    # Wait until all the workers are ready.
    pthread_barrier_wait(&sch.barrier)
    while 1:
        # Wait until a queue becomes available.
        sem_wait(&sch.num_free_queues)
        # If the scheduler is done there is nothing to do anymore.
        if sch.is_done:
            return <void*> 0
        # Pop or steal a queue.
        queue = worker.get_queue()
        with wlocked queue:
            # Do one task on the queue.
            queue.activate()
            if queue.is_empty():
                # Mark the empty queue as not assigned to any worker.
                del queue.worker
                # Decrement the number of non-completed queues.
                if sch.num_pending_queues.fetch_sub(1) == 1:
                    # Signal that there are no more queues.
                    sem_post(&sch.done)
                # Discard the empty queue and continue the main loop.
                continue
        pthread_mutex_lock(&worker.lock)
        # The queue is not empty: reinsert it in this worker's queues.
        worker.queues.push_back(queue)
        # Signal that the queue is available.
        sem_post(&sch.num_free_queues)
        pthread_mutex_lock(&worker.lock)


cdef cypclass Worker nolock:
    deque[SequentialMailBox] queues
    Scheduler scheduler
    pthread_t thread
    pthread_mutex_t lock

    Worker __new__(alloc, Scheduler scheduler):
        instance = alloc()
        instance.scheduler = scheduler
        pthread_mutex_init(&instance.lock, NULL)
        if not pthread_create(&instance.thread, NULL, worker_function, <void *> instance):
            return instance
        printf("pthread_create() failed\n")

    __dealloc__(self):
        pthread_mutex_destroy(&self.lock)

    SequentialMailBox get_queue(self):
        # Get the next queue in the worker's list or steal one.
        pthread_mutex_lock(&self.lock)
        if not self.queues.empty():
            queue = self.queues.front()
            self.queues.pop_front()
            pthread_mutex_unlock(&self.lock)
            return queue
        pthread_mutex_unlock(&self.lock)
        return self.steal_queue()

    SequentialMailBox steal_queue(self):
        # Steal a queue from another worker:
        # - inspect each worker in order starting at a random offset
        # - skip this worker and any worker with an empty queue list
        # - return the last queue of the first worker with a non-empty list
        cdef int i, index, num_workers, random_offset
        sch = self.scheduler
        num_workers = <int> sch.workers.size()
        random_offset = rand() % num_workers
        for i in range(num_workers):
            index = (i + random_offset) % num_workers
            victim = sch.workers[index]
            if victim is self:
                continue
            pthread_mutex_lock(&victim.lock)
            if not victim.queues.empty():
                stolen_queue = victim.queues.back()
                victim.queues.pop_back()
                stolen_queue.worker = self
                pthread_mutex_unlock(&victim.lock)
                return stolen_queue
            pthread_mutex_unlock(&victim.lock)
        return NULL

    int join(self):
        # Join the worker thread.
        return pthread_join(self.thread, NULL)


cdef cypclass Scheduler nolock:
    vector[Worker] workers
    pthread_barrier_t barrier
    sem_t num_free_queues
    atomic[int] num_pending_queues
    sem_t done
    volatile bint is_done

    __init__(self, int num_workers=0):
        if num_workers == 0: num_workers = sysconf(_SC_NPROCESSORS_ONLN)
        sem_init(&self.num_free_queues, 0, 0)
        sem_init(&self.done, 0, 0)
        self.num_pending_queues.store(0)
        if pthread_barrier_init(&self.barrier, NULL, num_workers + 1):
            printf("Could not allocate memory for the thread barrier\n")
            # Signal that no work will be done.
            sem_post(&self.done)
            return
        self.is_done = False
        self.workers.reserve(num_workers)
        for i in range(num_workers):
            worker = Worker(self)
            if worker is NULL:
                # Signal that no work will be done.
                sem_post(&self.done)
                return
            self.workers.push_back(worker)
        # Wait until all the worker threads are ready.
        pthread_barrier_wait(&self.barrier)

    __dealloc__(self):
        pthread_barrier_destroy(&self.barrier)
        sem_destroy(&self.num_free_queues)
        sem_destroy(&self.done)

    void post_queue(self, SequentialMailBox queue):
        # Add a queue to the first worker.
        main_worker = self.workers[0]
        pthread_mutex_lock(&main_worker.lock)
        queue.worker = main_worker
        main_worker.queues.push_back(queue)
        pthread_mutex_unlock(&main_worker.lock)
        # Increment the number of non-completed queues.
        self.num_pending_queues.fetch_add(1)
        # Signal that a queue is available.
        sem_post(&self.num_free_queues)

    void finish(self):
        # Wait until there is no more work.
        sem_wait(&self.done)
        # Signal the worker threads that there is no more work.
        self.is_done = True
        # Pretend that there are new queues to wake up the workers.
        for worker in self.workers:
            sem_post(&self.num_free_queues)
        # Clear the workers to break reference cycles.
        self.workers.clear()


cdef cypclass SequentialMailBox(ActhonQueueInterface):
    deque[ActhonMessageInterface] messages
    Scheduler scheduler
    Worker worker

    __init__(self, Scheduler scheduler):
        self.scheduler = scheduler
        self.worker = NULL

    bint is_empty(const self):
        return self.messages.empty()

    void push(self, ActhonMessageInterface message):
        # Add a task to the queue.
        self.messages.push_back(message)
        if message._sync_method is not NULL:
            message._sync_method.insertActivity(message)
        # If not worker is already assigned this queue
        # register it with the scheduler.
        if self.worker is NULL:
            self.scheduler.post_queue(self)

    bint activate(self):
        # Try to process the first message in the queue.
        cdef bint one_message_processed
        if self.messages.empty():
            return False
        next_message = self.messages.front()
        self.messages.pop_front()
        one_message_processed = next_message.activate()
        if one_message_processed:
            if next_message._sync_method is not NULL:
                next_message._sync_method.removeActivity(next_message)
        else:
            printf("Pushed front message to back :/\n")
            self.messages.push_back(next_message)
        return one_message_processed


cdef cypclass BatchMailBox(SequentialMailBox):
    bint activate(self):
        # Process as many messages as possible.
        while not self.messages.empty():
            next_message = self.messages.front()
            self.messages.pop_front()
            if not next_message.activate():
                printf("Pushed front message to back :/\n")
                self.messages.push_back(next_message)
                return False
            if next_message._sync_method is not NULL:
                next_message._sync_method.removeActivity(next_message)
        return True


cdef inline ActhonResultInterface NullResult() nogil:
    return NULL
